@@ -2,6 +2,9 @@
 #
 # Copyright © 2013-2014 Kimmo Parviainen-Jalanko.
 #
+"""
+IKEv2 Payloads as specified in RFC 5996 sections 3.2 - 3.16
+"""
 from enum import IntEnum
 from functools import reduce
 import ipaddress
@@ -17,6 +20,7 @@ from ike.util.prf import prf
 from .proposal import Proposal
 from .util.conv import to_bytes
 
+
 PRIVATE_KEY_PEM = 'tests/private_key.pem'
 
 __author__ = 'kimvais'
@@ -26,8 +30,7 @@ logger = logging.getLogger(__name__)
 
 class Type(IntEnum):
     """
-    Payload types as defined in
-    https://www.iana.org/assignments/ikev2-parameters/ikev2-parameters.xhtml#ikev2-parameters-2
+    Payload types from `IANA <https://www.iana.org/assignments/ikev2-parameters/ikev2-parameters.xhtml#ikev2-parameters-2>`_
     """
     no_next_payload = 0
     SA = 33
@@ -57,6 +60,9 @@ class Type(IntEnum):
 
 
 class _IkePayload(object):
+    """
+    Generic payload header `RFC5996 Section 3.2 <https://tools.ietf.org/html/rfc5996#section-3.2>`_
+    """
     _type = None
 
     def __init__(self, data=None, next_payload=Type.no_next_payload, critical=False):
@@ -65,7 +71,7 @@ class _IkePayload(object):
             next_payload, self.flags, self.length = const.PAYLOAD_HEADER.unpack(
                 data[:const.PAYLOAD_HEADER.size])
             self.next_payload = Type(next_payload)
-            self.parse(data[const.PAYLOAD_HEADER.size:])
+            self.parse(data[const.PAYLOAD_HEADER.size:self.length])
         else:
             self.next_payload = next_payload
             self.length = 0
@@ -96,6 +102,9 @@ class _IkePayload(object):
 
 
 class SA(_IkePayload):
+    """
+    `Security Association Payload <https://tools.ietf.org/html/rfc5996#section-3.3>`_
+    """
     def __init__(self, data=None, proposals=None, next_payload=Type.no_next_payload,
                  critical=False):
         super(SA, self).__init__(data, next_payload, critical)
@@ -115,9 +124,10 @@ class SA(_IkePayload):
                     ('AUTH_HMAC_SHA2_256_128',)
                 ])
             ]
+            self.spi = self.proposals[0].spi
         else:
             self.proposals = proposals
-        self.spi = self.proposals[0].spi
+            self.spi = self.proposals[0].spi
 
     def __bytes__(self):
         ret = list()
@@ -132,12 +142,19 @@ class SA(_IkePayload):
         last = False
         while not last:
             proposal = Proposal(data=data)
+            if proposal.spi:
+                logger.debug("Setting SPI to: {}".format(proposal.spi))
+                self.spi = proposal.spi
             self.proposals.append(proposal)
             last = proposal.last
             data = data[proposal.len:]
+        logger.debug("got {} proposals".format(len(self.proposals)))
 
 
 class KE(_IkePayload):
+    """
+    `Key Exchange Payload <https://tools.ietf.org/html/rfc5996#section-3.4>`_
+    """
     def parse(self, data):
         self.group, _ = struct.unpack('!2H', data[4:8])
         self.kex_data = data[const.PAYLOAD_HEADER.size + 4:self.length]
@@ -156,6 +173,9 @@ class KE(_IkePayload):
 
 
 class Nonce(_IkePayload):
+    """
+    `Nonce Payload <https://tools.ietf.org/html/rfc5996#section-3.9>`_
+    """
     def parse(self, data):
         self._data = data[const.PAYLOAD_HEADER.size:self.length]
 
@@ -173,6 +193,9 @@ class Nonce(_IkePayload):
 
 
 class Notify(_IkePayload):
+    """
+    `Notify Payload <https://tools.ietf.org/html/rfc5996#section-3.10>`_
+    """
     def __init__(self, notify_type=None, data=None, next_payload=Type.no_next_payload, critical=False):
         # TODO; Implement generation of notifications with data
         assert notify_type or data
@@ -206,6 +229,7 @@ class Notify(_IkePayload):
 
 class _TS(_IkePayload):
     """
+    `Traffic Selector Payload <https://tools.ietf.org/html/rfc5996#section-3.13>`_
     Single IPv4 address:port
     """
     def __init__(self, addr=None, data=None, next_payload=Type.no_next_payload, critical=False):
@@ -214,6 +238,7 @@ class _TS(_IkePayload):
         if addr:
             ip = int(ipaddress.IPv4Address(addr[0]))
             port = addr[1]
+            port = 0
 
             # Generate traffic selector
             selector = struct.pack("!2BH2H2I", 7, 0, 16, port, port, ip, ip)
@@ -221,14 +246,23 @@ class _TS(_IkePayload):
             self.length = len(self._data) + 4
 
 class TSi(_TS):
+    """
+    `Traffic Selector Payload <https://tools.ietf.org/html/rfc5996#section-3.13>`_ for initiator
+    """
     pass
 
 
 class TSr(_TS):
+    """
+    `Traffic Selector Payload <https://tools.ietf.org/html/rfc5996#section-3.13>`_ for responder
+    """
     pass
 
 
 class IDi(_IkePayload):
+    """
+    `Identification Payload <https://tools.ietf.org/html/rfc5996#section-3.5>`_ for initiator
+    """
     def __init__(self, data=None, next_payload=Type.no_next_payload, critical=False):
         super().__init__(data, next_payload, critical)
         EMAIL = b'test@77.fi'
@@ -237,10 +271,16 @@ class IDi(_IkePayload):
 
 
 class IDr(_IkePayload):
+    """
+    `Identification Payload <https://tools.ietf.org/html/rfc5996#section-3.5>`_ for responder
+    """
     pass
 
 
 class AUTH(_IkePayload):
+    """
+    `Authentication Payload <https://tools.ietf.org/html/rfc5996#section-3.8>`_
+    """
     def __init__(self, signed_octets=None, data=None, next_payload=Type.no_next_payload, critical=False):
         assert signed_octets or data
         super().__init__(data, next_payload, critical)
@@ -260,10 +300,13 @@ class AUTH(_IkePayload):
                 authentication_data = b''
                 raise AssertionError("Unsupported authentication method")
             self.length = 8 + len(authentication_data)
-            self._data = struct.pack("!B3x", authentication_type) + authentication_data
+            self._data = struct.pack(const.AUTH_HEADER, authentication_type) + authentication_data
 
 
 class SK(_IkePayload):
+    """
+    `Encrypted Payload <https://tools.ietf.org/html/rfc5996#section-3.14>`_
+    """
     def __init__(self, data=None, next_payload=Type.no_next_payload, critical=False, iv=None, ciphertext=None):
         assert data or (iv and ciphertext)
         super().__init__(data, next_payload, critical)
